@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2023 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2024 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -63,33 +63,30 @@ struct S3Fx {
 
 S3Fx::S3Fx() {
   // Create bucket
-  bool exists;
-  REQUIRE(s3_.is_bucket(S3_BUCKET, &exists).ok());
+  bool exists = s3_.is_bucket(S3_BUCKET);
   if (exists)
-    REQUIRE(s3_.remove_bucket(S3_BUCKET).ok());
+    REQUIRE_NOTHROW(s3_.remove_bucket(S3_BUCKET));
 
-  REQUIRE(s3_.is_bucket(S3_BUCKET, &exists).ok());
+  exists = s3_.is_bucket(S3_BUCKET);
   REQUIRE(!exists);
-  REQUIRE(s3_.create_bucket(S3_BUCKET).ok());
+  REQUIRE_NOTHROW(s3_.create_bucket(S3_BUCKET));
 
   // Check if bucket is empty
-  bool is_empty;
-  REQUIRE(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+  bool is_empty = s3_.is_empty_bucket(S3_BUCKET);
   CHECK(is_empty);
 }
 
 S3Fx::~S3Fx() {
   // Empty bucket
-  bool is_empty;
-  CHECK(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+  bool is_empty = s3_.is_empty_bucket(S3_BUCKET);
   if (!is_empty) {
-    CHECK(s3_.empty_bucket(S3_BUCKET).ok());
-    CHECK(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+    CHECK_NOTHROW(s3_.empty_bucket(S3_BUCKET));
+    is_empty = s3_.is_empty_bucket(S3_BUCKET);
     CHECK(is_empty);
   }
 
   // Delete bucket
-  CHECK(s3_.remove_bucket(S3_BUCKET).ok());
+  CHECK_NOTHROW(s3_.remove_bucket(S3_BUCKET));
   CHECK(s3_.disconnect().ok());
 }
 
@@ -105,202 +102,6 @@ Config S3Fx::set_config_params() {
   return config;
 }
 
-TEST_CASE_METHOD(S3Fx, "Test S3 filesystem, file management", "[s3]") {
-  /* Create the following file hierarchy:
-   *
-   * TEST_DIR/dir/subdir/file1
-   * TEST_DIR/dir/subdir/file2
-   * TEST_DIR/dir/file3
-   * TEST_DIR/file4
-   * TEST_DIR/file5
-   */
-  auto dir = TEST_DIR + "dir/";
-  auto dir2 = TEST_DIR + "dir2/";
-  auto subdir = dir + "subdir/";
-  auto file1 = subdir + "file1";
-  auto file2 = subdir + "file2";
-  auto file3 = dir + "file3";
-  auto file4 = TEST_DIR + "file4";
-  auto file5 = TEST_DIR + "file5";
-  auto file6 = TEST_DIR + "file6";
-
-  // Check that bucket is empty
-  bool is_empty;
-  CHECK(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
-  CHECK(is_empty);
-
-  // Continue building the hierarchy
-  bool exists = false;
-  CHECK(s3_.touch(URI(file1)).ok());
-  CHECK(s3_.is_object(URI(file1), &exists).ok());
-  CHECK(exists);
-  CHECK(s3_.touch(URI(file2)).ok());
-  CHECK(s3_.is_object(URI(file2), &exists).ok());
-  CHECK(exists);
-  CHECK(s3_.touch(URI(file3)).ok());
-  CHECK(s3_.is_object(URI(file3), &exists).ok());
-  CHECK(exists);
-  CHECK(s3_.touch(URI(file4)).ok());
-  CHECK(s3_.is_object(URI(file4), &exists).ok());
-  CHECK(exists);
-  CHECK(s3_.touch(URI(file5)).ok());
-  CHECK(s3_.is_object(URI(file5), &exists).ok());
-  CHECK(exists);
-
-  // Check that the bucket is not empty
-  CHECK(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
-  CHECK(!is_empty);
-
-  // Check invalid file
-  CHECK(s3_.is_object(URI(TEST_DIR + "foo"), &exists).ok());
-  CHECK(!exists);
-
-  // List with prefix
-  std::vector<std::string> paths;
-  CHECK(s3_.ls(URI(TEST_DIR), &paths).ok());
-  CHECK(paths.size() == 3);
-  paths.clear();
-  CHECK(s3_.ls(URI(dir), &paths).ok());
-  CHECK(paths.size() == 2);
-  paths.clear();
-  CHECK(s3_.ls(URI(subdir), &paths).ok());
-  CHECK(paths.size() == 2);
-  paths.clear();
-  CHECK(s3_.ls(S3_BUCKET, &paths, "").ok());  // No delimiter
-  CHECK(paths.size() == 5);
-
-  // Check if a directory exists
-  bool is_dir = false;
-  CHECK(s3_.is_dir(URI(file1), &is_dir).ok());
-  CHECK(!is_dir);  // Not a dir
-  CHECK(s3_.is_dir(URI(file4), &is_dir).ok());
-  CHECK(!is_dir);  // Not a dir
-  CHECK(s3_.is_dir(URI(dir), &is_dir).ok());
-  CHECK(is_dir);  // This is viewed as a dir
-  CHECK(s3_.is_dir(URI(TEST_DIR + "dir"), &is_dir).ok());
-  CHECK(is_dir);  // This is viewed as a dir
-
-  // ls_with_sizes
-  std::string s = "abcdef";
-  CHECK(s3_.write(URI(file3), s.data(), s.size()).ok());
-  CHECK(s3_.flush_object(URI(file3)).ok());
-
-  auto&& [status, rv] = s3_.ls_with_sizes(URI(dir));
-  auto children = *rv;
-  REQUIRE(status.ok());
-
-  REQUIRE(children.size() == 2);
-  CHECK(children[0].path().native() == file3);
-  CHECK(children[1].path().native() == subdir.substr(0, subdir.size() - 1));
-
-  CHECK(children[0].file_size() == s.size());
-  // Directories don't get a size
-  CHECK(children[1].file_size() == 0);
-
-  // Move file
-  CHECK(s3_.move_object(URI(file5), URI(file6)).ok());
-  CHECK(s3_.is_object(URI(file5), &exists).ok());
-  CHECK(!exists);
-  CHECK(s3_.is_object(URI(file6), &exists).ok());
-  CHECK(exists);
-  paths.clear();
-  CHECK(s3_.ls(S3_BUCKET, &paths, "").ok());  // No delimiter
-  CHECK(paths.size() == 5);
-
-  // Move directory
-  CHECK(s3_.move_dir(URI(dir), URI(dir2)).ok());
-  CHECK(s3_.is_dir(URI(dir), &is_dir).ok());
-  CHECK(!is_dir);
-  CHECK(s3_.is_dir(URI(dir2), &is_dir).ok());
-  CHECK(is_dir);
-  paths.clear();
-  CHECK(s3_.ls(S3_BUCKET, &paths, "").ok());  // No delimiter
-  CHECK(paths.size() == 5);
-
-  // Remove files
-  CHECK(s3_.remove_object(URI(file4)).ok());
-  CHECK(s3_.is_object(URI(file4), &exists).ok());
-  CHECK(!exists);
-
-  // Remove directories
-  CHECK(s3_.remove_dir(URI(dir2)).ok());
-  CHECK(s3_.is_object(URI(file1), &exists).ok());
-  CHECK(!exists);
-  CHECK(s3_.is_object(URI(file2), &exists).ok());
-  CHECK(!exists);
-  CHECK(s3_.is_object(URI(file3), &exists).ok());
-  CHECK(!exists);
-}
-
-TEST_CASE_METHOD(S3Fx, "Test S3 filesystem, file I/O", "[s3]") {
-  // Prepare buffers
-  uint64_t buffer_size = 5 * 1024 * 1024;
-  auto write_buffer = new char[buffer_size];
-  for (uint64_t i = 0; i < buffer_size; i++)
-    write_buffer[i] = (char)('a' + (i % 26));
-  uint64_t buffer_size_small = 1024 * 1024;
-  auto write_buffer_small = new char[buffer_size_small];
-  for (uint64_t i = 0; i < buffer_size_small; i++)
-    write_buffer_small[i] = (char)('a' + (i % 26));
-
-  // Write to two files
-  auto largefile = TEST_DIR + "largefile";
-  CHECK(s3_.write(URI(largefile), write_buffer, buffer_size).ok());
-  CHECK(s3_.write(URI(largefile), write_buffer_small, buffer_size_small).ok());
-  auto smallfile = TEST_DIR + "smallfile";
-  CHECK(s3_.write(URI(smallfile), write_buffer_small, buffer_size_small).ok());
-
-  // Before flushing, the files do not exist
-  bool exists = false;
-  CHECK(s3_.is_object(URI(largefile), &exists).ok());
-  CHECK(!exists);
-  CHECK(s3_.is_object(URI(smallfile), &exists).ok());
-  CHECK(!exists);
-
-  // Flush the files
-  CHECK(s3_.flush_object(URI(largefile)).ok());
-  CHECK(s3_.flush_object(URI(smallfile)).ok());
-
-  // After flushing, the files exist
-  CHECK(s3_.is_object(URI(largefile), &exists).ok());
-  CHECK(exists);
-  CHECK(s3_.is_object(URI(smallfile), &exists).ok());
-  CHECK(exists);
-
-  // Get file sizes
-  uint64_t nbytes = 0;
-  CHECK(s3_.object_size(URI(largefile), &nbytes).ok());
-  CHECK(nbytes == (buffer_size + buffer_size_small));
-  CHECK(s3_.object_size(URI(smallfile), &nbytes).ok());
-  CHECK(nbytes == buffer_size_small);
-
-  // Read from the beginning
-  auto read_buffer = new char[26];
-  uint64_t bytes_read = 0;
-  CHECK(s3_.read(URI(largefile), 0, read_buffer, 26, 0, &bytes_read).ok());
-  CHECK(26 == bytes_read);
-  bool allok = true;
-  for (int i = 0; i < 26; i++) {
-    if (read_buffer[i] != static_cast<char>('a' + i)) {
-      allok = false;
-      break;
-    }
-  }
-  CHECK(allok);
-
-  // Read from a different offset
-  CHECK(s3_.read(URI(largefile), 11, read_buffer, 26, 0, &bytes_read).ok());
-  CHECK(26 == bytes_read);
-  allok = true;
-  for (int i = 0; i < 26; i++) {
-    if (read_buffer[i] != static_cast<char>('a' + (i + 11) % 26)) {
-      allok = false;
-      break;
-    }
-  }
-  CHECK(allok);
-}
-
 TEST_CASE_METHOD(S3Fx, "Test S3 multiupload abort path", "[s3]") {
   // Prepare a large buffer
   uint64_t buffer_size = 100 * 1024 * 1024;
@@ -315,7 +116,7 @@ TEST_CASE_METHOD(S3Fx, "Test S3 multiupload abort path", "[s3]") {
     // Write one large file, the write will fail
     auto largefile =
         TEST_DIR + "failed_largefile_" + std::to_string(nth_failure);
-    CHECK(!s3_.write(URI(largefile), write_buffer, buffer_size).ok());
+    CHECK_THROWS(s3_.write(URI(largefile), write_buffer, buffer_size));
 
     // Before flushing, the file does not exist
     bool exists = false;
@@ -323,7 +124,7 @@ TEST_CASE_METHOD(S3Fx, "Test S3 multiupload abort path", "[s3]") {
     CHECK(!exists);
 
     // Flush the file
-    CHECK(s3_.flush_object(URI(largefile)).ok());
+    CHECK(!s3_.flush_object(URI(largefile)).ok());
 
     // After flushing, the file does not exist
     CHECK(s3_.is_object(URI(largefile), &exists).ok());
@@ -367,18 +168,16 @@ TEST_CASE_METHOD(S3Fx, "Test S3 use BucketCannedACL", "[s3]") {
     tiledb::sm::S3 s3_{&g_helper_stats, &thread_pool_, config};
 
     // Create bucket
-    bool exists;
-    REQUIRE(s3_.is_bucket(S3_BUCKET, &exists).ok());
+    bool exists = s3_.is_bucket(S3_BUCKET);
     if (exists)
-      REQUIRE(s3_.remove_bucket(S3_BUCKET).ok());
+      REQUIRE_NOTHROW(s3_.remove_bucket(S3_BUCKET));
 
-    REQUIRE(s3_.is_bucket(S3_BUCKET, &exists).ok());
+    exists = s3_.is_bucket(S3_BUCKET);
     REQUIRE(!exists);
-    REQUIRE(s3_.create_bucket(S3_BUCKET).ok());
+    REQUIRE_NOTHROW(s3_.create_bucket(S3_BUCKET));
 
     // Check if bucket is empty
-    bool is_empty;
-    REQUIRE(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+    bool is_empty = s3_.is_empty_bucket(S3_BUCKET);
     CHECK(is_empty);
 
     CHECK(s3_.disconnect().ok());
@@ -422,30 +221,29 @@ TEST_CASE_METHOD(S3Fx, "Test S3 use Bucket/Object CannedACL", "[s3]") {
     auto file6 = TEST_DIR + "file6";
 
     // Check that bucket is empty
-    bool is_empty;
-    CHECK(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+    bool is_empty = s3_.is_empty_bucket(S3_BUCKET);
     CHECK(is_empty);
 
     // Continue building the hierarchy
     bool exists = false;
-    CHECK(s3_.touch(URI(file1)).ok());
+    CHECK_NOTHROW(s3_.touch(URI(file1)));
     CHECK(s3_.is_object(URI(file1), &exists).ok());
     CHECK(exists);
-    CHECK(s3_.touch(URI(file2)).ok());
+    CHECK_NOTHROW(s3_.touch(URI(file2)));
     CHECK(s3_.is_object(URI(file2), &exists).ok());
     CHECK(exists);
-    CHECK(s3_.touch(URI(file3)).ok());
+    CHECK_NOTHROW(s3_.touch(URI(file3)));
     CHECK(s3_.is_object(URI(file3), &exists).ok());
     CHECK(exists);
-    CHECK(s3_.touch(URI(file4)).ok());
+    CHECK_NOTHROW(s3_.touch(URI(file4)));
     CHECK(s3_.is_object(URI(file4), &exists).ok());
     CHECK(exists);
-    CHECK(s3_.touch(URI(file5)).ok());
+    CHECK_NOTHROW(s3_.touch(URI(file5)));
     CHECK(s3_.is_object(URI(file5), &exists).ok());
     CHECK(exists);
 
     // Check that the bucket is not empty
-    CHECK(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+    is_empty = s3_.is_empty_bucket(S3_BUCKET);
     CHECK(!is_empty);
 
     // Check invalid file
@@ -488,7 +286,7 @@ TEST_CASE_METHOD(S3Fx, "Test S3 use Bucket/Object CannedACL", "[s3]") {
     CHECK(paths.size() == 5);
 
     // Move directory
-    CHECK(s3_.move_dir(URI(dir), URI(dir2)).ok());
+    CHECK_NOTHROW(s3_.move_dir(URI(dir), URI(dir2)));
     CHECK(s3_.is_dir(URI(dir), &is_dir).ok());
     CHECK(!is_dir);
     CHECK(s3_.is_dir(URI(dir2), &is_dir).ok());
@@ -503,7 +301,7 @@ TEST_CASE_METHOD(S3Fx, "Test S3 use Bucket/Object CannedACL", "[s3]") {
     CHECK(!exists);
 
     // Remove directories
-    CHECK(s3_.remove_dir(URI(dir2)).ok());
+    CHECK_NOTHROW(s3_.remove_dir(URI(dir2)));
     CHECK(s3_.is_object(URI(file1), &exists).ok());
     CHECK(!exists);
     CHECK(s3_.is_object(URI(file2), &exists).ok());
@@ -521,18 +319,16 @@ TEST_CASE_METHOD(S3Fx, "Test S3 use Bucket/Object CannedACL", "[s3]") {
     tiledb::sm::S3 s3_{&g_helper_stats, &thread_pool_, config};
 
     // Create bucket
-    bool exists;
-    REQUIRE(s3_.is_bucket(S3_BUCKET, &exists).ok());
+    bool exists = s3_.is_bucket(S3_BUCKET);
     if (exists)
-      REQUIRE(s3_.remove_bucket(S3_BUCKET).ok());
+      REQUIRE_NOTHROW(s3_.remove_bucket(S3_BUCKET));
 
-    REQUIRE(s3_.is_bucket(S3_BUCKET, &exists).ok());
+    exists = s3_.is_bucket(S3_BUCKET);
     REQUIRE(!exists);
-    REQUIRE(s3_.create_bucket(S3_BUCKET).ok());
+    REQUIRE_NOTHROW(s3_.create_bucket(S3_BUCKET));
 
     // Check if bucket is empty
-    bool is_empty;
-    REQUIRE(s3_.is_empty_bucket(S3_BUCKET, &is_empty).ok());
+    bool is_empty = s3_.is_empty_bucket(S3_BUCKET);
     CHECK(is_empty);
 
     exercise_object_canned_acl();
