@@ -5,7 +5,7 @@
  *
  * The MIT License
  *
- * @copyright Copyright (c) 2017-2022 TileDB, Inc.
+ * @copyright Copyright (c) 2017-2024 TileDB, Inc.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy
  * of this software and associated documentation files (the "Software"), to deal
@@ -30,8 +30,9 @@
  * This file implements class DenseReader.
  */
 
+#include "tiledb/sm/query/readers/dense_reader.h"
 #include "tiledb/common/logger.h"
-
+#include "tiledb/common/memory_tracker.h"
 #include "tiledb/sm/array/array.h"
 #include "tiledb/sm/array_schema/array_schema.h"
 #include "tiledb/sm/array_schema/dimension.h"
@@ -40,7 +41,6 @@
 #include "tiledb/sm/misc/utils.h"
 #include "tiledb/sm/query/legacy/cell_slab_iter.h"
 #include "tiledb/sm/query/query_macros.h"
-#include "tiledb/sm/query/readers/dense_reader.h"
 #include "tiledb/sm/query/readers/filtered_data.h"
 #include "tiledb/sm/query/readers/result_tile.h"
 #include "tiledb/sm/stats/global_stats.h"
@@ -54,8 +54,7 @@ using namespace tiledb;
 using namespace tiledb::common;
 using namespace tiledb::sm::stats;
 
-namespace tiledb {
-namespace sm {
+namespace tiledb::sm {
 
 class DenseReaderStatusException : public StatusException {
  public:
@@ -73,8 +72,7 @@ DenseReader::DenseReader(
     shared_ptr<Logger> logger,
     StrategyParams& params,
     bool remote_query)
-    : ReaderBase(stats, logger->clone("DenseReader", ++logger_id_), params)
-    , array_memory_tracker_(params.memory_tracker()) {
+    : ReaderBase(stats, logger->clone("DenseReader", ++logger_id_), params) {
   elements_mode_ = false;
 
   // Sanity checks.
@@ -232,7 +230,7 @@ Status DenseReader::dense_read() {
   // For easy reference.
   const auto dim_num = array_schema_.dim_num();
   auto& subarray = read_state_.partitioner_.current();
-  RETURN_NOT_OK(subarray.compute_tile_coords<DimType>());
+  subarray.compute_tile_coords<DimType>();
   const auto& domain{array_schema_.domain()};
 
   // Cache tile extents.
@@ -420,27 +418,28 @@ Status DenseReader::dense_read() {
     // clear the memory. Also, a name in names might not be in the user buffers
     // so we might skip the copy but still clear the memory.
     for (auto& name : names) {
-      if (name == constants::coords || array_schema_.is_dim(name)) {
-        continue;
-      }
-
-      // Get the tiles to load for this attribute.
-      auto result_tiles = result_tiles_to_load(
-          name,
-          qc_loaded_attr_names_set_,
-          subarray,
-          t_start,
-          t_end,
-          result_space_tiles,
-          tile_subarrays);
-
-      std::vector<FilteredData> filtered_data;
-
-      // Read and unfilter tiles.
+      shared_ptr<std::list<FilteredData>> filtered_data;
+      std::vector<ResultTile*> result_tiles;
       bool validity_only = null_count_aggregate_only(name);
-      std::vector<ReaderBase::NameToLoad> to_load;
-      to_load.emplace_back(name, validity_only);
-      filtered_data = std::move(read_attribute_tiles(to_load, result_tiles));
+      bool dense_dim = name == constants::coords || array_schema_.is_dim(name);
+
+      if (!dense_dim) {
+        // Get the tiles to load for this attribute.
+        result_tiles = result_tiles_to_load(
+            name,
+            qc_loaded_attr_names_set_,
+            subarray,
+            t_start,
+            t_end,
+            result_space_tiles,
+            tile_subarrays);
+
+        // Read and unfilter tiles.
+        std::vector<ReaderBase::NameToLoad> to_load;
+        to_load.emplace_back(name, validity_only);
+        filtered_data = make_shared<std::list<FilteredData>>(
+            read_attribute_tiles(to_load, result_tiles));
+      }
 
       if (compute_task.valid()) {
         RETURN_NOT_OK(storage_manager_->compute_tp()->wait(compute_task));
@@ -449,40 +448,43 @@ Status DenseReader::dense_read() {
         }
       }
 
-      compute_task = storage_manager_->compute_tp()->execute(
-          [&,
-           filtered_data = std::move(filtered_data),
-           name,
-           validity_only,
-           t_start,
-           t_end,
-           subarray_start_cell,
-           subarray_end_cell,
-           num_range_threads,
-           result_tiles]() {
-            // Unfilter tiles.
-            RETURN_NOT_OK(unfilter_tiles(name, validity_only, result_tiles));
+      compute_task =
+          storage_manager_->compute_tp()->execute([&,
+                                                   filtered_data,
+                                                   dense_dim,
+                                                   name,
+                                                   validity_only,
+                                                   t_start,
+                                                   t_end,
+                                                   subarray_start_cell,
+                                                   subarray_end_cell,
+                                                   num_range_threads,
+                                                   result_tiles]() {
+            if (!dense_dim) {
+              // Unfilter tiles.
+              RETURN_NOT_OK(unfilter_tiles(name, validity_only, result_tiles));
 
-            // Only copy names that are present in the user buffers.
-            if (buffers_.count(name) != 0) {
-              // Copy attribute data to users buffers.
-              auto& var_buffer_size = var_buffer_sizes[name];
-              status = copy_attribute<DimType, OffType>(
-                  name,
-                  tile_extents,
-                  subarray,
-                  t_start,
-                  t_end,
-                  subarray_start_cell,
-                  subarray_end_cell,
-                  tile_subarrays,
-                  tile_offsets,
-                  var_buffer_size,
-                  range_info,
-                  result_space_tiles,
-                  qc_result,
-                  num_range_threads);
-              RETURN_CANCEL_OR_ERROR(status);
+              // Only copy names that are present in the user buffers.
+              if (buffers_.count(name) != 0) {
+                // Copy attribute data to users buffers.
+                auto& var_buffer_size = var_buffer_sizes[name];
+                status = copy_attribute<DimType, OffType>(
+                    name,
+                    tile_extents,
+                    subarray,
+                    t_start,
+                    t_end,
+                    subarray_start_cell,
+                    subarray_end_cell,
+                    tile_subarrays,
+                    tile_offsets,
+                    var_buffer_size,
+                    range_info,
+                    result_space_tiles,
+                    qc_result,
+                    num_range_threads);
+                RETURN_CANCEL_OR_ERROR(status);
+              }
             }
 
             if (aggregates_.count(name) != 0) {
@@ -501,7 +503,9 @@ Status DenseReader::dense_read() {
               RETURN_CANCEL_OR_ERROR(status);
             }
 
-            clear_tiles(name, result_tiles);
+            if (!dense_dim) {
+              clear_tiles(name, result_tiles);
+            }
 
             return Status::Ok();
           });
@@ -509,9 +513,12 @@ Status DenseReader::dense_read() {
 
     // Process count aggregates.
     if (aggregates_.count(constants::count_of_rows) != 0) {
-      auto buff{make_aggregate_buffer(
+      DimType unused = 0;
+      auto buff{make_aggregate_buffer<DimType>(
           false,
           false,
+          false,
+          unused,
           0,
           subarray_start_cell,
           subarray_end_cell,
@@ -912,139 +919,141 @@ Status DenseReader::apply_query_condition(
         tile_subarrays);
 
     // Read and unfilter query condition attributes.
-    std::vector<FilteredData> filtered_data = read_attribute_tiles(
-        NameToLoad::from_string_vec(qc_names), result_tiles);
+    shared_ptr<std::list<FilteredData>> filtered_data =
+        make_shared<std::list<FilteredData>>(read_attribute_tiles(
+            NameToLoad::from_string_vec(qc_names), result_tiles));
 
     if (compute_task.valid()) {
       RETURN_NOT_OK(storage_manager_->compute_tp()->wait(compute_task));
     }
 
-    compute_task = storage_manager_->compute_tp()->execute(
-        [&,
-         filtered_data = std::move(filtered_data),
-         qc_names,
-         t_start,
-         t_end,
-         num_range_threads,
-         result_tiles]() {
-          // For easy reference.
-          const auto& tile_coords = subarray.tile_coords();
-          const auto dim_num = array_schema_.dim_num();
-          auto stride = array_schema_.domain().stride<DimType>(layout_);
-          const auto cell_order = array_schema_.cell_order();
-          const auto global_order = layout_ == Layout::GLOBAL_ORDER;
+    compute_task = storage_manager_->compute_tp()->execute([&,
+                                                            filtered_data,
+                                                            qc_names,
+                                                            t_start,
+                                                            t_end,
+                                                            num_range_threads,
+                                                            result_tiles]() {
+      // For easy reference.
+      const auto& tile_coords = subarray.tile_coords();
+      const auto dim_num = array_schema_.dim_num();
+      auto stride = array_schema_.domain().stride<DimType>(layout_);
+      const auto cell_order = array_schema_.cell_order();
+      const auto global_order = layout_ == Layout::GLOBAL_ORDER;
 
-          // Unfilter tiles.
-          for (auto& name : qc_names) {
-            RETURN_NOT_OK(unfilter_tiles(name, false, result_tiles));
-          }
+      // Unfilter tiles.
+      for (auto& name : qc_names) {
+        RETURN_NOT_OK(unfilter_tiles(name, false, result_tiles));
+      }
 
-          if (stride == UINT64_MAX) {
-            stride = 1;
-          }
+      if (stride == UINT64_MAX) {
+        stride = 1;
+      }
 
-          // Process all tiles in parallel.
-          auto status = parallel_for_2d(
-              storage_manager_->compute_tp(),
-              t_start,
-              t_end,
-              0,
-              num_range_threads,
-              [&](uint64_t t, uint64_t range_thread_idx) {
-                // Find out result space tile and tile subarray.
-                const DimType* tc = (DimType*)&tile_coords[t][0];
-                auto& result_space_tile = result_space_tiles.at(tc);
+      // Process all tiles in parallel.
+      auto status = parallel_for_2d(
+          storage_manager_->compute_tp(),
+          t_start,
+          t_end,
+          0,
+          num_range_threads,
+          [&](uint64_t t, uint64_t range_thread_idx) {
+            // Find out result space tile and tile subarray.
+            const DimType* tc = (DimType*)&tile_coords[t][0];
+            auto& result_space_tile = result_space_tiles.at(tc);
 
-                // Iterate over all coordinates, retrieved in cell slab.
-                const auto& frag_domains = result_space_tile.frag_domains();
-                TileCellSlabIter<DimType> iter(
-                    range_thread_idx,
-                    num_range_threads,
-                    subarray,
-                    tile_subarrays[t],
-                    tile_extents,
-                    result_space_tile.start_coords(),
-                    range_info,
-                    cell_order);
+            // Iterate over all coordinates, retrieved in cell slab.
+            const auto& frag_domains = result_space_tile.frag_domains();
+            TileCellSlabIter<DimType> iter(
+                range_thread_idx,
+                num_range_threads,
+                subarray,
+                tile_subarrays[t],
+                tile_extents,
+                result_space_tile.start_coords(),
+                range_info,
+                cell_order);
 
-                // Compute cell offset and destination pointer.
-                uint64_t cell_offset =
-                    global_order ? tile_offsets[t] + iter.global_offset() : 0;
-                auto dest_ptr = qc_result.data() + cell_offset;
+            // Compute cell offset and destination pointer.
+            uint64_t cell_offset =
+                global_order ? tile_offsets[t] + iter.global_offset() : 0;
+            auto dest_ptr = qc_result.data() + cell_offset;
 
-                while (!iter.end()) {
-                  // Compute destination pointer for row/col major orders.
-                  if (!global_order) {
-                    cell_offset = iter.dest_offset_row_col();
-                    dest_ptr = qc_result.data() + cell_offset;
-                  }
+            while (!iter.end()) {
+              // Compute destination pointer for row/col major orders.
+              if (!global_order) {
+                cell_offset = iter.dest_offset_row_col();
+                dest_ptr = qc_result.data() + cell_offset;
+              }
 
-                  for (int32_t i =
-                           static_cast<int32_t>(frag_domains.size()) - 1;
-                       i >= 0;
-                       --i) {
-                    // If the cell slab overlaps this fragment domain range,
-                    // apply clause.
-                    auto&& [overlaps, start, end] = cell_slab_overlaps_range(
-                        dim_num,
-                        frag_domains[i].domain(),
-                        iter.cell_slab_coords(),
-                        iter.cell_slab_length());
-                    if (overlaps) {
-                      // Re-initialize the bitmap to 1 in case of overlapping
-                      // domains.
-                      if (i != static_cast<int32_t>(frag_domains.size()) - 1) {
-                        for (uint64_t c = start; c <= end; c++) {
-                          dest_ptr[c] = 1;
-                        }
-                      }
-
-                      RETURN_NOT_OK(condition_->apply_dense(
-                          *(fragment_metadata_[frag_domains[i].fid()]
-                                ->array_schema()
-                                .get()),
-                          result_space_tile.result_tile(frag_domains[i].fid()),
-                          start,
-                          end - start + 1,
-                          iter.pos_in_tile(),
-                          stride,
-                          iter.cell_slab_coords().data(),
-                          dest_ptr));
-
-                      // If any cell doesn't match the query condition, signal
-                      // it in the space tile.
-                      for (uint64_t c = start; c <= end; c++) {
-                        if (dest_ptr[c] == 0) {
-                          result_space_tile.set_qc_filtered_results();
-                          break;
-                        }
-                      }
+              for (int32_t i = static_cast<int32_t>(frag_domains.size()) - 1;
+                   i >= 0;
+                   --i) {
+                // If the cell slab overlaps this fragment domain range,
+                // apply clause.
+                auto&& [overlaps, start, end] = cell_slab_overlaps_range(
+                    dim_num,
+                    frag_domains[i].domain(),
+                    iter.cell_slab_coords(),
+                    iter.cell_slab_length());
+                if (overlaps) {
+                  // Re-initialize the bitmap to 1 in case of overlapping
+                  // domains.
+                  if (i != static_cast<int32_t>(frag_domains.size()) - 1) {
+                    for (uint64_t c = start; c <= end; c++) {
+                      dest_ptr[c] = 1;
                     }
                   }
 
-                  // Adjust the destination pointers for global order.
-                  if (global_order) {
-                    dest_ptr += iter.cell_slab_length();
+                  QueryCondition::Params params(
+                      query_memory_tracker_,
+                      *(fragment_metadata_[frag_domains[i].fid()]
+                            ->array_schema()
+                            .get()));
+                  RETURN_NOT_OK(condition_->apply_dense(
+                      params,
+                      result_space_tile.result_tile(frag_domains[i].fid()),
+                      start,
+                      end - start + 1,
+                      iter.pos_in_tile(),
+                      stride,
+                      iter.cell_slab_coords().data(),
+                      dest_ptr));
+
+                  // If any cell doesn't match the query condition, signal
+                  // it in the space tile.
+                  for (uint64_t c = start; c <= end; c++) {
+                    if (dest_ptr[c] == 0) {
+                      result_space_tile.set_qc_filtered_results();
+                      break;
+                    }
                   }
-
-                  ++iter;
                 }
+              }
 
-                return Status::Ok();
-              });
-          RETURN_NOT_OK(status);
+              // Adjust the destination pointers for global order.
+              if (global_order) {
+                dest_ptr += iter.cell_slab_length();
+              }
 
-          // For `qc_coords_mode` just fill in the coordinates and skip
-          // attribute
-          // processing.
-          if (qc_coords_mode_) {
-            for (auto& name : qc_names) {
-              clear_tiles(name, result_tiles);
+              ++iter;
             }
-          }
 
-          return Status::Ok();
-        });
+            return Status::Ok();
+          });
+      RETURN_NOT_OK(status);
+
+      // For `qc_coords_mode` just fill in the coordinates and skip
+      // attribute
+      // processing.
+      if (qc_coords_mode_) {
+        for (auto& name : qc_names) {
+          clear_tiles(name, result_tiles);
+        }
+      }
+
+      return Status::Ok();
+    });
   }
 
   return Status::Ok();
@@ -1265,9 +1274,12 @@ Status DenseReader::copy_attribute(
   return Status::Ok();
 }
 
+template <class DimType>
 AggregateBuffer DenseReader::make_aggregate_buffer(
     const bool var_sized,
     const bool nullable,
+    const bool is_dim,
+    DimType& dim_val,
     const uint64_t cell_size,
     const uint64_t min_cell,
     const uint64_t max_cell,
@@ -1287,6 +1299,8 @@ AggregateBuffer DenseReader::make_aggregate_buffer(
             std::make_optional(
                 tile_tuple->validity_tile().data_as<uint8_t>() + min_cell) :
             nullopt;
+  } else if (is_dim) {
+    fixed_data = &dim_val;
   }
 
   return AggregateBuffer(
@@ -1864,14 +1878,28 @@ Status DenseReader::aggregate_tiles(
   const auto cell_order = array_schema_.cell_order();
   auto stride = array_schema_.domain().stride<DimType>(layout_);
   const auto& frag_domains = result_space_tile.frag_domains();
+  const auto is_dim = array_schema_.is_dim(name);
   const auto attribute = array_schema_.attribute(name);
   const auto var_size = array_schema_.var_size(name);
-  const auto nullable = attribute->nullable();
+  const auto nullable = !is_dim && attribute->nullable();
   const auto cell_size = var_size ? constants::cell_var_offset_size :
                                     array_schema_.cell_size(name);
   auto& aggregates = aggregates_[name];
   const bool validity_only = null_count_aggregate_only(name);
 
+  // Get the dimension index.
+  unsigned dim_idx = 0;
+  if (is_dim) {
+    dim_idx = array_schema_.domain().get_dimension_index(name);
+  }
+
+  const bool is_slab_dim = is_dim && (cell_order == sm::Layout::ROW_MAJOR) ?
+                               (dim_idx == dim_num - 1) :
+                               (dim_idx == 0);
+  const bool is_col_dim = is_dim && (cell_order == sm::Layout::ROW_MAJOR) ?
+                              (dim_idx == 0) :
+                              (dim_idx == dim_num - 1);
+
   // Cache tile tuples.
   std::vector<ResultTile::TileTuple*> tile_tuples(frag_domains.size());
   for (uint32_t fd = 0; fd < frag_domains.size(); ++fd) {
@@ -1915,7 +1943,7 @@ Status DenseReader::aggregate_tiles(
       // If the cell slab overlaps this fragment domain range, copy data.
       bool overlaps = false;
       uint64_t start = 0, end = 0;
-      if (tile_tuples[fd] != nullptr) {
+      if (is_dim || tile_tuples[fd] != nullptr) {
         auto&& [o, s, e] = cell_slab_overlaps_range(
             dim_num,
             frag_domains[fd].domain(),
@@ -1928,11 +1956,18 @@ Status DenseReader::aggregate_tiles(
       if (overlaps) {
         // If the subarray and tile are in the same order, aggregate the
         // whole slab.
+        DimType dim_val = 0;
         if (stride == 1) {
+          if (is_dim) {
+            dim_val = iter.cell_slab_coords()[dim_idx] + is_slab_dim * start;
+          }
+
           // Compute aggregate.
-          AggregateBuffer aggregate_buffer{make_aggregate_buffer(
+          AggregateBuffer aggregate_buffer{make_aggregate_buffer<DimType>(
               var_size & !validity_only,
               nullable,
+              is_dim,
+              dim_val,
               cell_size,
               iter.pos_in_tile() + start,
               iter.pos_in_tile() + end + 1,
@@ -1944,11 +1979,18 @@ Status DenseReader::aggregate_tiles(
         } else {
           // Go cell by cell.
           for (uint64_t i = 0; i < end - start + 1; ++i) {
+            if (is_dim) {
+              dim_val =
+                  iter.cell_slab_coords()[dim_idx] + is_col_dim * (i + start);
+            }
+
             // Compute aggregate.
             auto start_cell = iter.pos_in_tile() + (start + i) * stride;
-            AggregateBuffer aggregate_buffer{make_aggregate_buffer(
+            AggregateBuffer aggregate_buffer{make_aggregate_buffer<DimType>(
                 var_size & !validity_only,
                 nullable,
+                is_dim,
+                dim_val,
                 cell_size,
                 start_cell,
                 start_cell + 1,
@@ -2257,5 +2299,4 @@ void DenseReader::fill_dense_coords_col_slab(
   }
 }
 
-}  // namespace sm
-}  // namespace tiledb
+}  // namespace tiledb::sm
