@@ -199,8 +199,7 @@ Status fragment_info_request_deserialize(
 std::tuple<Status, std::optional<SingleFragmentInfo>>
 single_fragment_info_from_capnp(
     const capnp::SingleFragmentInfo::Reader& single_frag_info_reader,
-    const std::unordered_map<std::string, shared_ptr<ArraySchema>>&
-        array_schemas) {
+    FragmentInfo* fragment_info) {
   // Get array schema name
   std::string schema_name;
   if (single_frag_info_reader.hasArraySchemaName()) {
@@ -213,8 +212,8 @@ single_fragment_info_from_capnp(
   }
 
   // Use the array schema name to find the corresponding array schema
-  auto schema = array_schemas.find(schema_name);
-  if (schema == array_schemas.end()) {
+  auto schema = fragment_info->array_schemas_all().find(schema_name);
+  if (schema == fragment_info->array_schemas_all().end()) {
     return {
         Status_SerializationError(
             "Could not find schema" + schema_name +
@@ -226,7 +225,10 @@ single_fragment_info_from_capnp(
   shared_ptr<FragmentMetadata> meta;
   if (single_frag_info_reader.hasMeta()) {
     auto frag_meta_reader = single_frag_info_reader.getMeta();
-    meta = make_shared<FragmentMetadata>(HERE());
+
+    auto memory_tracker = fragment_info->resources()->create_memory_tracker();
+    meta = make_shared<FragmentMetadata>(
+        HERE(), fragment_info->resources(), memory_tracker);
     auto st =
         fragment_metadata_from_capnp(schema->second, frag_meta_reader, meta);
   } else {
@@ -274,16 +276,16 @@ Status single_fragment_info_to_capnp(
 Status fragment_info_from_capnp(
     const capnp::FragmentInfo::Reader& fragment_info_reader,
     const URI& array_uri,
-    FragmentInfo* fragment_info) {
+    FragmentInfo* fragment_info,
+    shared_ptr<MemoryTracker> memory_tracker) {
   // Get array_schema_latest from capnp
   if (fragment_info_reader.hasArraySchemaLatest()) {
     auto array_schema_latest_reader =
         fragment_info_reader.getArraySchemaLatest();
-    auto array_schema_latest{
-        array_schema_from_capnp(array_schema_latest_reader, array_uri)};
-    array_schema_latest.set_array_uri(array_uri);
-    fragment_info->array_schema_latest() =
-        make_shared<ArraySchema>(HERE(), array_schema_latest);
+    auto array_schema_latest{array_schema_from_capnp(
+        array_schema_latest_reader, array_uri, memory_tracker)};
+    array_schema_latest->set_array_uri(array_uri);
+    fragment_info->array_schema_latest() = array_schema_latest;
   }
 
   // Get array_schemas_all from capnp
@@ -293,14 +295,13 @@ Status fragment_info_from_capnp(
     if (all_schemas_reader.hasEntries()) {
       auto entries = fragment_info_reader.getArraySchemasAll().getEntries();
       for (auto array_schema_build : entries) {
-        auto schema{
-            array_schema_from_capnp(array_schema_build.getValue(), array_uri)};
-        schema.set_array_uri(array_uri);
+        auto schema{array_schema_from_capnp(
+            array_schema_build.getValue(), array_uri, memory_tracker)};
+        schema->set_array_uri(array_uri);
         auto key = std::string_view{
             array_schema_build.getKey().cStr(),
             array_schema_build.getKey().size()};
-        fragment_info->array_schemas_all()[std::string{key}] =
-            make_shared<ArraySchema>(HERE(), schema);
+        fragment_info->array_schemas_all()[std::string{key}] = schema;
       }
     }
   }
@@ -312,7 +313,7 @@ Status fragment_info_from_capnp(
         fragment_info_list_reader.size());
     for (auto single_frag_info_reader : fragment_info_list_reader) {
       auto&& [st, single_frag_info] = single_fragment_info_from_capnp(
-          single_frag_info_reader, fragment_info->array_schemas_all());
+          single_frag_info_reader, fragment_info);
       RETURN_NOT_OK(st);
       fragment_info->single_fragment_info_vec().emplace_back(
           single_frag_info.value());
@@ -451,7 +452,8 @@ Status fragment_info_deserialize(
     FragmentInfo* fragment_info,
     SerializationType serialize_type,
     const URI& uri,
-    const Buffer& serialized_buffer) {
+    const Buffer& serialized_buffer,
+    shared_ptr<MemoryTracker> memory_tracker) {
   if (fragment_info == nullptr)
     return LOG_STATUS(
         Status_SerializationError("Error deserializing fragment info; null "
@@ -469,7 +471,8 @@ Status fragment_info_deserialize(
         auto reader = builder.asReader();
 
         // Deserialize
-        RETURN_NOT_OK(fragment_info_from_capnp(reader, uri, fragment_info));
+        RETURN_NOT_OK(fragment_info_from_capnp(
+            reader, uri, fragment_info, memory_tracker));
         break;
       }
       case SerializationType::CAPNP: {
@@ -491,7 +494,8 @@ Status fragment_info_deserialize(
         auto reader = msg_reader.getRoot<capnp::FragmentInfo>();
 
         // Deserialize
-        RETURN_NOT_OK(fragment_info_from_capnp(reader, uri, fragment_info));
+        RETURN_NOT_OK(fragment_info_from_capnp(
+            reader, uri, fragment_info, memory_tracker));
         break;
       }
       default: {
@@ -522,7 +526,11 @@ Status fragment_info_serialize(
 }
 
 Status fragment_info_deserialize(
-    FragmentInfo*, SerializationType, const URI&, const Buffer&) {
+    FragmentInfo*,
+    SerializationType,
+    const URI&,
+    const Buffer&,
+    shared_ptr<MemoryTracker>) {
   return LOG_STATUS(Status_SerializationError(
       "Cannot deserialize; serialization not enabled."));
 }
